"""
"""

import collections
import string
import yaml

from importlib.resources import read_text
from .. import resources


def make_color_key(name):
    """Normalizes a color name to a lowercase string with
    no embedded white spaces and no embedded punctuation.
    """
    tt = str.maketrans({k: "" for k in string.punctuation + string.whitespace})
    return name.translate(tt).strip().casefold()


COLORS_BY_NAME = collections.ChainMap()
COLORS_BY_NAME.maps.pop()

for catalog_file in resources.color_catalog_files:
    catalog = {}
    for k, v in yaml.safe_load(read_text(resources, catalog_file)).items():
        catalog.setdefault(make_color_key(k), tuple(v))
    COLORS_BY_NAME.maps.append(catalog)


NAMES_BY_COLOR = {v: k for k, v in COLORS_BY_NAME.items()}
# this clobbers some alternate spellings/names
# aqua =/= cyan
# grey =/= gray in substrings of names

ALL_NAMES_BY_COLOR = {}
for k, v in COLORS_BY_NAME.items():
    ALL_NAMES_BY_COLOR.setdefault(v, []).append(k)



""" python syntax highlighting tool

"""

import click
import io
import sys
import yaml


from . import VERSION
from .render import PythonRender
from .style_book import StyleBook


# Fun comment here
#


@click.command()
@click.argument(
    "input-file", type=click.File(mode="rb"), default=sys.stdin,
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(),
    default=None,
    help="Creates a PNG with the supplied path.",
)
@click.option(
    "-l",
    "--start-line",
    default=0,
    help="Line number to begin display.",
    show_default=True,
)
@click.option(
    "-n",
    "--line-count",
    default=10,
    help="Number of lines to display.",
    show_default=True,
)
@click.option(
    "-N",
    "--no-line-numbers",
    is_flag=True,
    default=False,
    help="Disable line numbers in output.",
    show_default=True,
)
@click.option(
    "-p",
    "--preview",
    is_flag=True,
    default=False,
    help="Previews output in window.",
    show_default=True,
)
@click.option(
    "-t",
    "--transparent",
    is_flag=True,
    default=False,
    help="Write output PNG with transparency.",
    show_default=True,
)
@click.option(
    "-s",
    "--style-name",
    type=click.STRING,
    default="default",
    help="Style to apply to input file.",
    show_default=True,
)
@click.option(
    "-f",
    "--font-name",
    type=click.STRING,
    default="courier",
    help="Font name.",
    show_default=True,
)
@click.option(
    "-S",
    "--font-size",
    type=click.INT,
    default=24,
    help="Font size",
    show_default=True,
)
@click.option(
    "-L",
    "--list-styles",
    is_flag=True,
    default=False,
    help="List available styles and exits.",
)
@click.option("-d", "--debug", is_flag=True, hidden=True, default=False)
@click.version_option(VERSION)
def pyliter_cli(
    input_file,
    output_file,
    start_line,
    line_count,
    no_line_numbers,
    preview,
    transparent,
    style_name,
    font_name,
    font_size,
    list_styles,
    debug,
):
    """Python syntax highlighting

    Renders syntax-highlighted text to PNG file and optional previews
    the render in a window before saving.

    If the optional output path is omitted, preview is enabled.

    """

    if list_styles:
        for style in StyleBook.available_styles():
            print(style)
        return

    try:
        stylebook = StyleBook.by_name(style_name)
    except FileNotFoundError:
        print(click.get_current_context().get_help())
        print(f"\nUnknown style: '{style_name}'\n")
        return

    if input_file == sys.stdin:
        input_file = io.BytesIO(sys.stdin.buffer.read())

    if not output_file:
        preview = True

    stylebook.default["font_name"] = font_name
    stylebook.default["font_size"] = font_size

    render = PythonRender(
        input_file,
        start_line,
        line_count,
        stylebook,
        preview,
        transparent,
        not no_line_numbers,
    )

    render.run(output_file)



color_catalog_files = [
    "css_colors.yaml",
    "crayola_bright.yaml",
    "crayola_colors.yaml",
    "crayola_fluorescent.yaml",
    "crayola_fragrance.yaml",
    "crayola_jewelry.yaml",
    "crayola_magic.yaml",
    "crayola_metallic.yaml",
    "crayola_scary.yaml",
    "crayola_silver.yaml",
]


__all__ = ["color_catalog_files"]



"""
"""

import yaml

from pathlib import Path
import importlib.resources
from . import resources
from .color import Color


class StyleBook(dict):
    @classmethod
    def available_styles(cls) -> list:
        """Return a list of available style books by name.
        """
        styles = []
        for filename in importlib.resources.contents(resources):
            path = Path(filename)
            if path.match("*_style.yaml"):
                styles.append(path.stem.replace("_style", ""))
        return styles

    @classmethod
    def by_name(cls, style_name: str):
        """Return a StyleBook initialized with the contents of the
        resource file identified by 'style_name'.
        """
        with importlib.resources.path(resources, f"{style_name}_style.yaml") as path:
            return cls.from_filename(path)

    @classmethod
    def from_filename(cls, filename: str):
        """Return a StyleBook initializes with the contents of the given filename.
        """
        path = Path(filename)
        return cls(yaml.safe_load(path.read_text()))

    def __init__(self, styles: dict):
        """
        :param dict styles: dictionary of dictionaries
        """

        self.update(styles)

        for category, attributes in self.items():
            for color_key in ["color", "background_color", "underline"]:
                try:
                    color_spec = attributes[color_key]
                    color = Color.from_any(color_spec)
                    attributes[color_key] = color.rgba
                except KeyError:
                    pass

    def get(self, key: str) -> dict:
        """Returns a dictionary for the given key.
        """
        return super().get(key, {})

    def __str__(self):
        """YAML formatted string.
        """
        return yaml.safe_dump(dict(self), sort_keys=False)

    def save(self, path):
        """Save the contents of this StyleBook to path in YAML format.
        """
        yaml.safe_dump(dict(self), open(path, "w"), sort_keys=False)

    @property
    def default(self):
        """Returns the DEFAULT style category dictionary.
        """
        return self.get("DEFAULT")

    @property
    def styles(self):
        """List of style category names.
        """
        return list(self.keys())



"""
"""


class Value0:
    minimum = 0
    maximum = 0

    @classmethod
    def from_hex_string(cls, hex_string):
        """
        """
        if hex_string.startswith("#"):
            hex_string = hex_string[1:]
        if hex_string.startswith("0x"):
            hex_string = hex_string[2:]
        return cls(int(hex_string, 16))

    @classmethod
    def from_scale(cls, scale_value):
        return cls(round(cls.maximum * scale_value))

    @classmethod
    def from_byte_offset(cls, value: int, byte_index=0):
        print("MAX", cls.maximum, cls.__name__)
        byte_value = (value >> (byte_index * 8)) & cls.maximum
        return cls(byte_value)

    def __init__(self, value: int):
        self.value = value

    def __repr__(self):
        return f"{self.__class__.__name__}(value=0x{self.hex})"

    def __str__(self) -> str:
        return self.hex

    @property
    def hex(self) -> str:
        return f"{self.value:02x}"

    @property
    def value(self):
        try:
            return self._value
        except AttributeError:
            pass
        self._value = 0
        return self._value

    @value.setter
    def value(self, value):
        self._value = min(self.maximum, max(value, self.minimum))

    @property
    def scale(self) -> float:
        return round(self.value / self.maximum, 2)


class Value8(Value0):
    maximum = (1 << 8) - 1


class Value16(Value0):
    maximum = (1 << 16) - 1



"""
"""

import click
import token as TOKEN
import builtins
import keyword

import tokenize
from dataclasses import dataclass, field
from pprint import pformat

_BUILTINS = dir(builtins)


KEYWORD = 1101
BUILTIN = 1102
DEFINITIONS = 1103
DOCSTRING = 1104
FNAME = 1105
CNAME = 1106
INAME = 1107
LVAL = 1108
HIDDEN = 1109
SELF = 1110


@dataclass
class ExtendedToken:
    type: int
    string: str
    start: tuple
    end: tuple
    line: str
    exact_type: int = 0
    category: str = field(default="", init=False, repr=True)
    attributes: dict = field(default_factory=dict, init=False, repr=False)

    @classmethod
    def from_file(cls, input_file, classifier=None) -> tuple:

        try:
            classifier = classifier()
        except TypeError:
            classifier = TokenClassifier()

        tokens = []

        for t in tokenize.tokenize(input_file.readline):
            tokens.append(cls(*t._asdict().values(), t.exact_type))
            classifier(tokens[-1], tokens[:-1])

        text = tokenize.untokenize(tokens).decode("utf-8")

        return tokens, text

    @classmethod
    def lines_from_file(
        cls, input_file, start_line: int, line_count: int, classifier=None
    ):

        tokens, text = cls.from_file(input_file, classifier)

        if line_count == -1:
            end_line = line_count
        else:
            end_line = start_line + line_count

        text = "\n".join(text.splitlines()[start_line:end_line])
        # TokenInfo.start is a tuple of (line number, offset), line number
        # starts at one instead of zero so subtract one to avoid a one-off bug.
        tokens = [t for t in tokens if start_line <= (t.start[0] - 1) < end_line]

        return tokens, text

    def __len__(self):
        return 4

    def __iter__(self):
        return iter([self.type, self.string, self.start, self.end, self.line])

    def __str__(self):
        return f"{self!r}\nA:\n{pformat(self.attributes)}"


class TokenClassifier:
    @property
    def builtins(self):
        try:
            return self._builtins
        except AttributeError:
            pass
        self._builtins = dir(builtins)
        return self._builtins

    @property
    def categories(self):
        try:
            return self._categories
        except AttributeError:
            pass

        self._categories = {}
        self._categories.update(tokenize.tok_name)
        self._categories[KEYWORD] = "KEYWORD"
        self._categories[BUILTIN] = "BUILTIN"
        self._categories[DEFINITIONS] = "DEFINITIONS"
        self._categories[DOCSTRING] = "DOCSTRING"
        self._categories[FNAME] = "FNAME"
        self._categories[CNAME] = "CNAME"
        self._categories[INAME] = "INAME"
        self._categories[LVAL] = "LVAL"
        self._categories[HIDDEN] = "HIDDEN"
        self._categories[SELF] = "SELF"

        return self._categories

    def __call__(self, token, prev_tokens):
        self.classify(token, prev_tokens)

    def classify(self, token, prev_tokens):

        while True:
            if keyword.iskeyword(token.string):
                token.exact_type = KEYWORD
                break

            if token.string in self.builtins:
                token.exact_type = BUILTIN
                break

            if token.string.startswith(('"""', "'''")):
                token.exact_type = DOCSTRING
                break

            if token.string.startswith("_"):
                token.exact_type = HIDDEN
                break

            if token.string == "self":
                token.exact_type = SELF
                break

            try:
                if prev_tokens[-1].string == "def":
                    token.exact_type = FNAME
                    break

                if prev_tokens[-1].string == "class":
                    token.exact_type = CNAME
                    break

                if prev_tokens[-1].string == "import":
                    token.exact_type = INAME
                    break
            except IndexError:
                pass
            break

        # need to backtrack for LVAL classification
        if token.type != TOKEN.OP:
            token.category = self.categories[token.exact_type]
        else:
            token.category = self.categories[token.type]



from .catalog import COLORS_BY_NAME, NAMES_BY_COLOR, make_color_key
from .exceptions import ColorNameNotFound
from .value import Value8


class Color:
    """A 32-bit RGBA based color with four 8-bit channels:

    - red
    - green
    - blue
    - alpha (transparency)

    Colors may be specified by their component values, chosen by name,
    hexadecimal string or integer representation, by HSL or HSV tuples.
    """

    @classmethod
    def from_any(cls, colorspec, tuple_type: str = None):

        try:
            return cls.from_value(colorspec)
        except ValueError:
            pass

        try:
            return cls.from_name(colorspec)
        except ColorNameNotFound:
            pass

        try:
            return cls.from_hex_string(colorspec)
        except ValueError:
            pass

        try:
            return cls(*colorspec)
        except Exception:
            raise TypeError(f"unable to decode colorspec {type(colorspec)}") from None

    @classmethod
    def from_name(cls, name: str, alpha: int = Value8.maximum):
        """Attempts to locate RGB values for the color name given.

        :param str name:
        :param int alpha: range [0,255], defaults to 255
        
        Raises:
        - ColorNameNotFound

        :param str name:
        :param int alpha:
        """
        try:
            r, g, b = COLORS_BY_NAME[make_color_key(name)]
        except KeyError:
            raise ColorNameNotFound(name) from None
        return cls(r, g, b, alpha)

    @classmethod
    def from_value(cls, value: int):
        """
        :param int value:
        """
        return cls.from_hex_string(f"{value:08x}")

    @classmethod
    def from_hex_string(cls, hex_string: str):
        """Returns a RGBA color constructed from a string composed of

        hexidecimal digits. Traditional hex string prefixes '#' or '0x'
        are ignored.

        Supported hex string formats:

        3 digit: RGB
        4 digit: RGBA
        6 digit: RRGGBB
        8 digit: RRGGBBAA

        R - red, G - green, B - blue, A - alpha

        If the alpha channel is not present (3 or 6 digits), it defaults
        to 0xff.

        Three and four digit hex values are scaled from 4-bit to 8-bit
        color.

        :param str hex_string:
        """

        try:
            if hex_string.startswith("#"):
                hex_string = hex_string[1:]
        except AttributeError:
            raise ValueError(f"str expected, not {type(hex_string)}")

        if hex_string.startswith("0x"):
            hex_string = hex_string[2:]

        hex_string_len = len(hex_string)

        if hex_string_len not in [3, 4, 6, 8]:
            raise ValueError(f"cannot decode '{hex_string}'")

        if hex_string_len in [3, 4]:

            def scale_8_to_16(value):
                return round(255 * (int(value, 16) / 15))

            return cls(*tuple(scale_8_to_16(v) for v in hex_string))

        byte_values = [hex_string[x : x + 2] for x in range(0, hex_string_len, 2)]

        return cls(*[int(v, 16) for v in byte_values])

    @classmethod
    def from_scale(cls, red: float, green: float, blue: float, alpha: float = 1.0):
        """
        :param float red: [0, 1]
        :param float green: [0, 1]
        :param float blue: [0, 1]
        :param float alpha: [0, 1]
        """
        color = cls(0, 0, 0, 0)
        color.red = Value8.from_scale(red)
        color.green = Value8.from_scale(green)
        color.blue = Value8.from_scale(blue)
        color.alpha = Value8.from_scale(alpha)
        return color

    @classmethod
    def from_hsl(
        cls, hue: float, saturation: float, lightness: float, alpha: float = 1.0
    ):
        """
        :param float hue: [0, 359] degrees
        :param float saturation: [0, 1]
        :param float lightness: [0, 1]
        :param float alpha: [0, 1]
        """
        hue %= 360
        saturation, lightness, alpha = map(
            lambda v: min(1.0, max(v, 0.0)), [saturation, lightness, alpha]
        )

        c = (1 - (abs((2 * lightness) - 1))) * saturation
        x = c * (1 - abs(((hue / 60) % 2) - 1))
        m = lightness - (c / 2)

        try:
            return cls._from_hcxma(hue, c, x, m, alpha, "hsl")
        except ValueError as error:
            raise error from None

    @classmethod
    def from_hsv(cls, hue: float, saturation: float, value: float, alpha: float = 1.0):
        """
        :param float hue: [0, 359] degrees
        :param float saturation: [0, 1]
        :param float value: [0, 1]
        :param float alpha: [0, 1]
        """
        hue %= 360
        saturation, value, alpha = map(
            lambda v: min(1.0, max(v, 0.0)), [saturation, value, alpha]
        )

        c = value * saturation
        x = c * (1 - abs(((hue / 60) % 2) - 1))
        m = value - c

        try:
            return cls._from_hcxma(hue, c, x, m, alpha, "hsv")
        except ValueError as error:
            raise error from None

    @classmethod
    def _from_hcxma(
        cls, hue: float, c: float, x: float, m: float, alpha: float, src: str
    ):
        """Common conversion from HSL/HSV to RGB.

        :param float hue: degrees constrained between 0 and 359
        :param float c: 
        :param float x:
        :param float m:
        :param float alpha: [0,1]
        :param str src: string identifying source conversion format ['hsv'|'hsl']
        """

        rgb = {
            0: (c + m, x + m, 0 + m),
            60: (x + m, c + m, 0 + m),
            120: (0 + m, c + m, x + m),
            180: (0 + m, x + m, c + m),
            240: (x + m, 0 + m, c + m),
            300: (c + m, 0 + m, x + m),
        }

        color = cls(0, 0, 0, 0)
        color.alpha = Value8.from_scale(alpha)

        for key, values in rgb.items():
            if key <= hue <= key + 60:
                color.red, color.green, color.blue = map(Value8.from_scale, values)
                break
        else:
            raise ValueError(f"problem converting {src} to rgb")

        return color

    @classmethod
    def from_cmyk(
        cls,
        cyan: float,
        magenta: float,
        yellow: float,
        black: float,
        alpha: float = 1.0,
    ):
        """
        :param float cyan: [0, 1]
        :param float magenta: [0, 1]
        :param float yellow: [0, 1]
        :param float black: [0, 1]
        :param float alpha: [0, 1]
        """

        r = (1.0 - cyan) * (1.0 - black)
        g = (1.0 - magenta) * (1.0 - black)
        b = (1.0 - yellow) * (1.0 - black)

        return cls.from_scale(r, g, b, alpha)

    def __init__(self, red: int, green: int, blue: int, alpha: int = Value8.maximum):
        """
        :param int red: [0, 255]
        :param int green: [0, 255]
        :param int blue: [0, 255]
        :param int alpha: [0, 255]
        """
        self.red = Value8(red)
        self.green = Value8(green)
        self.blue = Value8(blue)
        self.alpha = Value8(alpha)

    def __repr__(self):

        return "{}(red={}, green={}, blue={}, alpha={})".format(
            self.__class__.__name__,
            self.red.value,
            self.green.value,
            self.blue.value,
            self.alpha.value,
        )

    def __eq__(self, other):
        return self.rgba == other.rgba

    def __lt__(self, other):
        return self.value < other.value

    def __le__(self, other):
        return self.value <= other.value

    def __gt__(self, other):
        return self.value > other.value

    def __ge__(self, other):
        return self.value >= other.value

    def __hash__(self):
        return self.value

    @property
    def name(self) -> str:
        """Color name if RGB tuple present in the catalog otherwise 'unnamed'."""
        return NAMES_BY_COLOR.get(self.rgb, "unnamed")

    @property
    def value(self) -> int:
        """A 32-bit integer color representation.
        """
        return int(self.hex[1:], 16)

    @property
    def hex(self) -> str:
        """An eight digit hex string prefixed with an octothorpe: #RRGGBBAA
        """
        return "#" + self.red.hex + self.green.hex + self.blue.hex + self.alpha.hex

    @property
    def rgb(self) -> tuple:
        """Three-tuple of integer values: (red, green, blue).
        """
        return (self.red.value, self.green.value, self.blue.value)

    @property
    def rgba(self) -> tuple:
        """Four-tuple of integer values: (red, green, blue, alpha).
        """
        return self.rgb + (self.alpha.value,)

    @property
    def rgb_f(self) -> tuple:
        """Three-tuple of float values: (red, green, blue).
        """
        return (self.red.scale, self.green.scale, self.blue.scale)

    @property
    def rgba_f(self) -> tuple:
        """Four-tuple of float values: (red, green, blue, alpha).
        """
        return self.rgb_f + (self.alpha.scale,)

    @property
    def hsl(self) -> tuple:
        """Three-tuple of float values: (hue, saturation, lightness)

        Hue is a float in [0, 359]
        Saturation is a float in [0,1]
        Lightness is a float in [0,1]
        """
        c_max = max(self.rgb_f)
        c_min = min(self.rgb_f)
        c_delta = c_max - c_min

        try:
            if c_max == self.red.scale:
                hue = ((self.green.scale - self.blue.scale) / c_delta) % 6
            if c_max == self.green.scale:
                hue = ((self.blue.scale - self.red.scale) / c_delta) + 2
            if c_max == self.blue.scale:
                hue = ((self.red.scale - self.green.scale) / c_delta) + 4
        except ZeroDivisionError:
            hue = 0

        hue *= 60
        lightness = (c_max + c_min) / 2

        saturation = 0.0 if c_delta == 0 else (c_delta / (1 - abs(2 * lightness - 1)))

        return (hue, saturation, lightness)

    @property
    def hsv(self) -> tuple:
        """Three-tuple of float values: (hue, saturation, value)

        Hue is a float in [0, 359]
        Saturation is a float in [0,1]
        Lightness is a float in [0,1]
        """
        return self.hsl[:2] + (max(self.rgb_f),)

    @property
    def cmyk(self) -> tuple:
        """Four-tuple of float values: (cyan, magenta, yellow, black)
        
        Cyan is a float in [0, 1]
        Magenta is a float in [0, 1]
        Yellow is a float in [0, 1]
        Black is a float in [0, 1]
        """

        k = 1 - max(self.rgb_f)
        try:
            c = (1 - self.red.scale - k) / (1 - k)
            m = (1 - self.green.scale - k) / (1 - k)
            y = (1 - self.blue.scale - k) / (1 - k)
        except ZeroDivisionError:
            c, m, y = 0.0, 0.0, 0.0
        return (c, m, y, k)

    def as_dict(self) -> dict:
        """Returns a dictionary representation of this class.
        """
        return {
            "name": self.name,
            "red": self.red.value,
            "green": self.green.value,
            "blue": self.blue.value,
            "alpha": self.alpha.value,
        }



""" python rendering pyglet Window
"""

import pyglet
from .document import PythonDocument
from .color import Color


class PythonRender(pyglet.window.Window):
    def __init__(
        self,
        fileobj,
        start_line,
        line_count,
        style_book: dict = None,
        preview: bool = False,
        transparent: bool = False,
        number_lines: bool = False,
    ):
        """Render syntax-highlighted Python code to a window using
        the supplied style_book.

        :param fileobj:     file-like object containing Python text
        :param start_line:  line to begin rendering
        :param line_count:  number of lines to render
        :param style_book:  dictionary of style dictionaries
        :param preview:     controls whether the render text is displayed
        :param transparent: render with a transparent background if true
        """

        self.preview = preview

        self.document = PythonDocument(
            fileobj, start_line, line_count, style_book, number_lines
        )

        width, height = self.document.dimensions

        self.layout = pyglet.text.layout.TextLayout(
            self.document, width, multiline=True
        )
        self.layout.x = self.margin
        self.layout.y = self.margin

        try:
            caption = fileobj.name
        except AttributeError:
            caption = "stdin"

        super().__init__(
            width=width + (self.margin * 2),
            height=height + (self.margin * 2),
            visible=preview,
            caption=caption,
        )

        bg = Color(*self.document.background_color)
        if transparent:
            bg.alpha.value = 0

        pyglet.gl.glEnable(pyglet.gl.GL_BLEND)
        pyglet.gl.glBlendFunc(pyglet.gl.GL_SRC_ALPHA, pyglet.gl.GL_ONE_MINUS_SRC_ALPHA)
        pyglet.gl.glClearColor(*bg.rgba_f)

        vwidth, vheight = self.get_viewport_size()
        if (self.width, self.height) != (vwidth, vheight):
            pyglet.gl.glMatrixMode(pyglet.gl.GL_PROJECTION)
            pyglet.gl.glLoadIdentity()
            pyglet.gl.glOrtho(0, vwidth, 0, vheight, -1, 1)
            pyglet.gl.glMatrixMode(pyglet.gl.GL_MODELVIEW)

    @property
    def margin(self):
        """Pixels between edges and layout.
        """
        return self.document.font.height

    def on_draw(self):
        """Clear the window and draw it's contents (again).
        """
        self.clear()
        self.layout.draw()

    def save(self, filename: str):
        """Saves the contents of this window to filename.
        """

        if not filename:
            return

        image = pyglet.image.get_buffer_manager().get_color_buffer()

        if (self.width, self.height) != self.get_viewport_size():
            image = image.get_region(0, 0, self.width, self.height)

        image.save(filename)

    def run(self, filename: str = None):
        """Start up a pyglet event loop if in preview mode, otherwise
        call this object's on_draw method, save buffer contents to a
        file and call it a day.
        """
        if not self.preview:
            self.on_draw()
            self.save(filename)
            return

        pyglet.app.run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyliter', 'pyliter.color', 'pyliter.resources']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0', 'pyglet>=1.4.8,<2.0.0', 'pyyaml>=5.2,<6.0']

entry_points = \
{'console_scripts': ['pyliter = pyliter.__main__:pyliter_cli']}

setup_kwargs = {
    'name': 'pyliter',
    'version': '1.3.0',
    'description': 'Generate color syntax highlighted PNG image from python source files.',
    'long_description': 'pyliter - Python syntax highlighting\n====================================\n\n``pyliter`` is a Python 3 command-line tool that generates PNG files\nfrom python source. \n\n\nFeatures\n--------\n\n- syntax highlighting\n- PNG files\n- preview mode\n- OpenGL rendering using pyglet\n\nInstall\n-------\n\n::\n\n   $ pip install pyliter\n\n\n::\n\n   $ pip install git+https://github.com/JnyJny/pyliter\n\n\nUsage\n-----\n\n::\n\n   $ pyliter --help\n   Usage: pyliter [OPTIONS] [INPUT_FILE]\n   \n     Python syntax highlighting\n   \n     Renders syntax-highlighted text to PNG file and optional previews the\n     render in a window before saving.\n   \n     If the optional output path is omitted, preview is enabled.\n   \n   Options:\n     -o, --output-file PATH    Creates a PNG with the supplied path.\n     -l, --start-line INTEGER  Line number to begin display.  [default: 0]\n     -n, --line-count INTEGER  Number of lines to display.  [default: 10]\n     -N, --no-line-numbers     Disable line numbers in output.  [default: False]\n     -p, --preview             Previews output in window.  [default: False]\n     -t, --transparent         Write output PNG with transparency.  [default:\n                               False]\n     -s, --style-name TEXT     Style to apply to input file.  [default: default]\n     -f, --font-name TEXT      Font name.  [default: courier]\n     -S, --font-size INTEGER   Font size  [default: 24]\n     -L, --list-styles         List available styles and exits.\n     --version                 Show the version and exit.\n     --help                    Show this message and exit.\n\n\nExample\n-------\n\n.. image:: https://github.com/JnyJny/pyliter/blob/master/examples/screenshot.png\n\t   :width: 400\n\t   :alt: Super Awesome PNG Screenshot\n\n \n',
    'author': 'jnyjny',
    'author_email': 'erik.oshaughnessy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/JnyJny/pyliter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pyglet

from .token import ExtendedToken as PythonTokens
from .token import DOCSTRING


class PythonDocument(pyglet.text.document.FormattedDocument):
    def __init__(
        self,
        fileobj,
        start: int = 0,
        line_count: int = -1,
        style_book: dict = None,
        number_lines: bool = False,
        debug: bool = False,
    ):

        # The Order of Operations:
        # 1. setup subclass-specific attributes
        # 2. call __init__ of superclass
        # 3. set the default document style
        # 4. style documents with token derived attributes
        # 5. insert line numbers if requested
        # 6. trim document to keep lines in range [start, start+line_count]

        self.style_book = style_book or {
            "DEFAULT": {
                "font_name": "Source Code Pro",
                "font_size": 18,
                "color": (255, 255, 255, 255),
                "background_color": (0, 0, 0, 255),
                "line_numbers": False,
            }
        }

        self.tokens, self.raw_text = PythonTokens.from_file(fileobj)

        super().__init__(self.raw_text.translate(str.maketrans({"{": "{{", "}": "}}"})))

        self.set_style(start=0, end=self.eot, attributes=self.style_book["DEFAULT"])

        self._apply_token_styles()

        if number_lines:
            self._insert_line_numbers()

        if debug:
            # EJO - styles each line with a green underscore at the front and a
            #       red underscore at the end
            for i, (b, e) in enumerate(self._line_extents):
                self.set_style(b, b + 1, attributes={"underline": (0, 255, 0, 255)})
                self.set_style(e - 1, e, attributes={"underline": (255, 0, 0, 255)})

        self._keep(start, line_count)

    @property
    def eot(self):
        """End Of Text: integer position of the last character in self.text.
        """
        return len(self.text) + 1

    def _keep(self, start, line_count) -> None:
        """Keep the lines in the range of [start, start+line_count].
        """
        self._trim_lines(0, start)
        off = 0 if start == 0 else 1
        self._trim_lines(line_count + off, len(self.text.splitlines()))

    def _trim_lines(self, from_line: int, to_line: int) -> None:
        """Removes the lines from self.text in the range of [from_line, to_line]
        """
        if from_line == to_line:
            return
        extents = self._line_extents
        self.delete_text(extents[from_line][0], extents[to_line - 1][1])

    @property
    def _line_extents(self) -> list:
        """A list of tuples, each tuple is the starting and ending
        index into self.text of a line. Line 0 is the first tuple and
        so on. Lines are defined as a substring ending in a newline
        character '\n'.  Note: line extents change after text insert
        and delete options, so don't hang onto this list between
        operations.
        """
        p = 0
        extents = []
        while True:
            try:
                e = self.text.index("\n", p)
                extents.append((p, e))
                p = e + 1
            except ValueError:
                e = self.eot
                extents.append((p, e))
                break
        return extents

    def _insert_line_numbers(self, color: tuple = None) -> None:
        """Adds line numbers to the beginning of each line
        starting at line zero. Line numbers are four digits
        right justified and rendered with the given color.

        :param tuple color: 4-tuple of integers [0,255]
        """

        color = color or (0xA0, 0xA0, 0xA0, 0xA0)  # dark gray

        p = 0
        line_num = 0
        attributes = {
            "color": color,
            "underline": None,
            "bold": None,
            "italic": None,
        }
        while True:
            try:
                self.insert_text(p, f"{line_num:>4d} ", attributes=attributes)
                p = self.text.index("\n", p) + 1
                line_num += 1
            except ValueError:
                break

    def _apply_token_styles(self, tokens: list = None, style_book: dict = None) -> None:
        """Using the token category as a key into the style book, look for
        the starting position of each token and apply the appropriate style.

        :param list tokens:
        :param dict style_book:
        """

        tokens = tokens or self.tokens
        style_book = style_book or self.style_book

        p = 0

        for token in tokens:
            if not token.category in style_book:
                continue

            if token.exact_type == DOCSTRING:
                try:
                    p = self.text.index('"""', p)
                except ValueError:
                    pass
                try:
                    e = self.text.index('"""', p + 1) + 3
                except ValueError:
                    e = self.eot
            else:
                try:
                    p = self.text.index(token.string, p)
                    e = p + len(token.string)
                except ValueError:
                    print("F", p, token.string, repr(token))
                    continue

            self.set_style(p, e, style_book[token.category])
            p = e

    @property
    def background_color(self) -> tuple:
        """Convenience accessor for self.style_book['DEFAULT']['background_color'].
        """
        return self.style_book["DEFAULT"]["background_color"]

    @property
    def color(self) -> tuple:
        """Convenience accessor for self.style_book['DEFAULT']['color'].
        """
        return self.style_book["DEFAULT"]["color"]

    @property
    def dimensions(self) -> tuple:
        """A 2-tuple of width and height expressed in pixels.
        """

        lines = self.text.splitlines()
        maxc = max(len(l) for l in lines)
        nlines = len(lines) + 1

        return (
            maxc * self.font.size,
            nlines * self.font.height,
        )

    @property
    def font(self):
        """Convenience accessor for self.get_font() whose side-effect
        is setting a "height" attribute on font calculated with the
        sum of the font's ascent and abs(descent) values.
        """
        try:
            return self._font
        except AttributeError:
            pass

        self._font = self.get_font(0)

        self._font.height = self._font.ascent - self._font.descent

        return self._font
